/*
 * This example was taken from the transpose benchmark in the CUDA SDK (v5.0).
 * This kernel shows that loop invariants can be very involved.
 */

// Try verifying this kernel
//    gpuverify --blockDim=[4,2] --gridDim=[2,2] -DWIDTH=8 -DHEIGHT=8 -DTILE_DIM=4 -DBLOCK_ROWS=2 transpose.cu

/*
 * Note if you wish to change the dimensions then you must meet the following
 * preconditions:
 * --blockDim should equal [TILE_DIM, BLOCK_ROWS]
 * --gridDim  should equal [WIDTH/TILE_DIM, HEIGHT/TILE_DIM]
 */

#include <cuda.h>
#define __odata_offset (__write_offset(odata)/sizeof(float))

#ifndef WIDTH
#error Must define WIDTH
#endif

#ifndef HEIGHT
#error Must define HEIGHT
#endif

#ifndef TILE_DIM
#error Must define TILE_DIM
#endif

#ifndef BLOCK_ROWS
#error Must define BLOCK_ROWS
#endif

__global__ void transpose(float *odata, float *idata, int width, int height, int nreps)
{
    __requires(__is_pow2(BLOCK_ROWS));
    __requires(__is_pow2(TILE_DIM));
    __requires(BLOCK_ROWS <= TILE_DIM);
    __requires(__mod_pow2(TILE_DIM, BLOCK_ROWS) == 0);
    __requires(__mod_pow2(width, TILE_DIM) == 0);
    __requires(__mod_pow2(height, TILE_DIM) == 0);
    __requires(width == WIDTH);
    __requires(height == HEIGHT);

    __assert(blockDim.x == TILE_DIM);
    __assert(blockDim.y == BLOCK_ROWS);
    __assert(width  == gridDim.x * TILE_DIM);
    __assert(height == gridDim.y * TILE_DIM);

    int xIndex = blockIdx.x * TILE_DIM + threadIdx.x;
    int yIndex = blockIdx.y * TILE_DIM + threadIdx.y;

    int index_in  = xIndex + width * yIndex;
    int index_out = yIndex + height * xIndex;
  
    for (int i=0;
        __invariant(__mod_pow2(i, BLOCK_ROWS) == 0),
        __invariant(0 <= i),
        __invariant(i <= TILE_DIM),
        __invariant(__write_implies(odata, (__mod_pow2(__odata_offset, (height*TILE_DIM)) / height) == threadIdx.x)),
        __invariant(__write_implies(odata, __mod_pow2(__odata_offset, BLOCK_ROWS) == threadIdx.y)),
        __invariant(__write_implies(odata, __odata_offset/(height*TILE_DIM) == blockIdx.x)),
        __invariant(__write_implies(odata, (__mod_pow2(__odata_offset, height) / TILE_DIM) == blockIdx.y)),
        i<TILE_DIM; i+=BLOCK_ROWS)
    {
        odata[index_out+i] = idata[index_in+i*width];
    }

#ifdef FORCE_FAIL
    __assert(false);
#endif
}
